{ config, pkgs, lib, ... }:

with lib;
let localPackages = pkgs.buildPackages;
in {
  options.genode = {
    core = {

      prefix = mkOption {
        type = types.str;
        example = "hw-pc-";
      };

      supportedSystems = mkOption {
        type = types.listOf types.str;
        example = [ "i686-genode" "x86_64-genode" ];
      };

      basePackages = mkOption { type = types.listOf types.package; };

    };

    boot = {

      kernel = mkOption {
        type = types.path;
        default = "${pkgs.genodePackages.bender}/bender";
      };

      initrd = mkOption {
        type = types.str;
        default = "${pkgs.genodePackages.bender}/bender";
        description = "Path to an image or a command-line arguments";
      };

      config = mkOption {
        type = types.path;
        description = ''
          Dhall boot configuration. See
          https://git.sr.ht/~ehmry/dhall-genode/tree/master/Boot/package.dhall
        '';
      };

      image = mkOption {
        type = types.path;
        description =
          "Boot image containing the base component binaries and configuration.";
      };

      romModules = mkOption {
        type = types.attrsOf types.path;
        description = "Attr set of initial ROM modules";
      };

    };

  };

  config = let
    initInputs = unique config.genode.init.inputs;

    addManifest = drv:
      drv // {
        manifest =
          localPackages.runCommand "${drv.name}.dhall" { inherit drv; } ''
            set -eu
            echo -n '[' >> $out
            find $drv/ -type f -printf ',{mapKey= "%f",mapValue="%p"}' >> $out
            ${if builtins.elem "lib" drv.outputs then
              ''
                find ${drv.lib}/ -type f -printf ',{mapKey= "%f",mapValue="%p"}' >> $out''
            else
              ""}
            echo -n ']' >> $out
          '';
      };

    mergeManifests = inputs:
      localPackages.writeTextFile {
        name = "manifest.dhall";
        text = with builtins;
          let
            f = head: input:
              if hasAttr "manifest" input then
                ''
                  ${head}, { mapKey = "${
                    lib.getName input
                  }", mapValue = ${input.manifest} }''
              else
                abort "${input.pname} does not have a manifest";
          in (foldl' f "[" inputs) + "]";
      };

  in {

    assertions = [{
      assertion = builtins.any (s: s == config.nixpkgs.system)
        config.genode.core.supportedSystems;
      message = "invalid Genode core for this system";
    }];

    genode.boot.config = let
      manifest = mergeManifests (map addManifest
        (config.genode.core.basePackages ++ [ config.system.build.storeTarball ]
          ++ (with pkgs.genodePackages; [ init vfs cached_fs_rom ])));
    in localPackages.runCommand "boot.dhall" { } ''
      cat > $out << EOF
      ${./store-wrapper.dhall}
      (${config.genode.init.config})
      $(stat --format '%s' ${config.system.build.storeTarball}/store.tar)
      ${config.system.build.storeManifest} ${manifest}
      EOF
    '';

    system.build.storeManifest = mergeManifests (map addManifest initInputs);

    system.build.storeTarball = localPackages.runCommand "store" { } ''
      mkdir -p $out
      tar cf "$out/store.tar" --absolute-names ${toString initInputs}
    '';

    system.build.initXml = pkgs.buildPackages.runCommand "init.xml" {
      nativeBuildInputs = with pkgs.buildPackages; [ dhall xorg.lndir ];
      DHALL_GENODE = "${pkgs.genodePackages.dhallGenode}/binary.dhall";
      BOOT_CONFIG = config.genode.boot.config;
    } ''
      export XDG_CACHE_HOME=$NIX_BUILD_TOP
      lndir -silent \
        ${pkgs.genodePackages.dhallGenode}/.cache \
        $XDG_CACHE_HOME
      dhall text <<< "(env:DHALL_GENODE).Init.render (env:BOOT_CONFIG).config" > $out
    '';

  };

}
